from messages.arduino.KvMessage import KvIntMessage, KvMessage, MessageModeType
from messages.core.Message import EmptyMessage
from core.core import Core, TestSubscriber, Node
from nav.calc.Calc import inRange
import time
import queue
import threading
from utils.TimeUtils import timeElapsed
from utils.TimeUtils import getCurrentTimeMillis
from messages.core.Message import ErrorMessage


class MessageExpanderNode(Node):

    def __init__(self, core: Core):
        super().__init__(core)
        self.rules = {}

    def addRule(self, topic, messageList):
        self.rules[topic] = messageList
        self.core.subscribe(topic, self.input)

    def input(self, message):
        if message.topic in self.rules:
            for message in self.rules[message.topic]:
                self.core.publish(message)


class ParamSyncRule:

    def __init__(self, configParam: str, responseTopic: str, message: KvMessage) -> None:
        self.configParam = configParam
        self.responseTopic = responseTopic
        self.message = message


class ParamSyncNode(Node):

    def __init__(self, core: Core):
        super().__init__(core)
        self.mapper = {}
        self.waitingMessage = None
        self.expectedValue = None
        self.tries = 5
        self.responseTimeout = 2000
        self.topicQueue = queue.SimpleQueue()
        self.alive = True
        self.t = threading.Thread(target=self.run)
        self.t.start()

    def run(self):
        while self.alive:
            if self.topicQueue.qsize() > 0:
                topic = self.topicQueue.get()
                for paramSyncRule in self.mapper[topic]:
                    res = self.waitForTries(paramSyncRule)
                    if not res:
                        self.core.publish(ErrorMessage("MUru01: Zarizeni neodpovida na nastavovani nebo vraceni parametru {}".format(str(self.waitingMessage))))
                        self.abortWaiting()
            else:
                time.sleep(1)

    def waitForTries(self, paramSyncRule: ParamSyncRule):
        for i in range(self.tries):
            self.process(paramSyncRule)
            time.sleep(1)
            if self.waitForResponse():
                return True
        return False

    def waitForResponse(self):
        startTime = getCurrentTimeMillis()
        while not timeElapsed(startTime, self.responseTimeout):
            if self.isWaiting() is False:
                return True
            time.sleep(0.1)
        return False

    def addMessageParam(self, topic, paramSyncRuleList: [ParamSyncRule]):
        self.mapper[topic] = paramSyncRuleList
        self.core.subscribe(topic, self.toProcess)

    def toProcess(self, message):
        self.topicQueue.put(message.topic)

    def process(self, paramSyncRule: ParamSyncRule):
        if paramSyncRule.message.mode == MessageModeType.GET.value:
            self.getRemoteValue(paramSyncRule)
            paramSyncRule.message.mode = MessageModeType.GET.value

        if paramSyncRule.message.mode == MessageModeType.SET.value:
            self.setRemoteValue(paramSyncRule)
            self.expectedValue = paramSyncRule.message.value
            self.getRemoteValue(paramSyncRule)
            paramSyncRule.message.mode = MessageModeType.SET.value

    def setRemoteValue(self, paramSyncRule: ParamSyncRule):
        output = self.core.getParam(paramSyncRule.configParam)
        paramSyncRule.message.value = output
        paramSyncRule.message.mode = MessageModeType.SET.value
        self.core.publish(paramSyncRule.message)

    def getRemoteValue(self, paramSyncRule: ParamSyncRule):
        paramSyncRule.message.mode = MessageModeType.GET.value
        self.waitingMessage = [paramSyncRule.message, paramSyncRule.configParam, paramSyncRule.responseTopic]
        self.core.subscribe(paramSyncRule.responseTopic, self.resonseRemoteValue)
        self.core.publish(paramSyncRule.message)


    def resonseRemoteValue(self, message: KvMessage):
        #print("Response: {}".format(message.value))
        if self.waitingMessage is None:
            return
        wm = self.waitingMessage[0]
        param = self.waitingMessage[1]
        responseTopic = self.waitingMessage[2]
        if responseTopic == message.topic and wm.key == message.key:
            if self.expectedValue is not None:
                if inRange(message.value, self.expectedValue, 0.1):
                    self.core.setParam(param, message.value)
                    self.waitingMessage = None
                    self.expectedValue = None
            else:
                self.core.setParam(param, message.value)
                self.waitingMessage = None
                self.expectedValue = None

    def isWaiting(self):
        return self.waitingMessage is not None

    def abortWaiting(self):
        self.waitingMessage = None
        self.expectedValue = None

    def stop(self):
        self.alive = False

    def isProcessed(self):
        return self.waitingMessage is None and self.topicQueue.qsize() == 0



if __name__ == '__main__':
    c = Core()
    tr = TestSubscriber(c, "/kvim", True)
    m = KvIntMessage(key=32, value=26)
    e = MessageExpanderNode(c)
    e.addRule("/top", m)
    c.publish(EmptyMessage("/top"))


class MessagesEqual:

    @staticmethod
    def getTopicMessageEqual(message, expectedMessage):
        return message.topic == expectedMessage.topic