from messages.core.EnginesControlMessage import EnginesControlMessage
from messages.EncoderMessage import EncoderMessage
from messages.GpsMessage import GpsMessage
from messages.socket.GamepadMessage import GamepadMessage
from messages.MagnetoMessage import MagnetoMessage
import nav.calc.Calc
from nav.EncoderInfoNode import EncoderInfoNode
from core.core import Core, TestSubscriber
from enum import Enum
import statistics
from utils.ExtractorUtils import extract
from engines.DifferentialEnginesControl import DifferentialEnginesControl
from messages.arduino.BridgeMessage import BridgeMessage, ImprovedBridgeMessage
from messages.core.Message import EmptyMessage, ValueMessage, KeyValueMessage
#from simple_pid import PID
from nav.calc.ivPid import PID
from engines.EnginesContol import EnginesControl
from utils.ConvertUtils import Button
from socketio.socketio import FlaskBridgeParamNode
from messages.arduino.KvMessage import KvMessage
from utils.StringUtils import topicContains
from nav.calc.Calc import AngleCalc


class DirectionMode(Enum):
    HOLD_ANGLE = 1
    OFF = 2
    GAMEPAD_INPUT = 3

class TurnDirection(Enum):
    LEFT = -1
    NONE = 0
    RIGHT = 1

DIRECTION_MODE_DICT = {
    1: DirectionMode.HOLD_ANGLE,
    2: DirectionMode.OFF,
    3: DirectionMode.GAMEPAD_INPUT
}


class SpeedMode(Enum):
    HOLD_SPEED = 1
    OFF = 2
    GAMEPAD_INPUT = 3
    SPEED_SLOW = 4
    SPEED_MEDIUM = 5
    SPEED_HIGH = 6
    REACH_DISTANCE = 7
    HOLD_SPEED_TICKS = 8
    DISTANCE_SPEED_TICKS = 9

SPEED_MODE_DICT = {
    1: SpeedMode.HOLD_SPEED,
    2: SpeedMode.OFF,
    3: SpeedMode.GAMEPAD_INPUT,
    4: SpeedMode.SPEED_SLOW,
    5: SpeedMode.SPEED_MEDIUM,
    6: SpeedMode.SPEED_HIGH,
    7: SpeedMode.REACH_DISTANCE,
    8: SpeedMode.HOLD_SPEED_TICKS,
    9: SpeedMode.DISTANCE_SPEED_TICKS
}


class SimpleEnginesController(EnginesControl):
    FOLLOW_SPEED = "/followCards/speed"
    OBSTACLE_SPEED = "/followCards/obstacleSpeed"
    STOP_MESSAGE = "/stop"
    PARAM_BEARING_TOPIC = "/simpleEnginesController/bearingTopic"

    def __init__(self, core: Core):
        super().__init__(core)
        self.bridgeMessage = ImprovedBridgeMessage(mode=1)
        self.manual = True
        self.speedGamepadOverride = True
        self.waitingForTasks = False
        self.obstacle = False
        self.pid = None
        self.heading = 0
        self.bearing = 0
        self.direction = 90
        self.speed = 90
        self.resetPid(None)
        self.atd = nav.calc.Calc.AckermannToDifferential()
        self.softOnBtn = Button()
        self.softOffBtn = Button()
        self.manualOnBtn = Button()
        self.manualOffBtn = Button()
        self.core.registerTopic("/controlEnginesPid/updatePid", EmptyMessage)
        self.core.subscribe("/controlEnginesPid/updatePid", self.pid)
        self.core.subscribe("/gamepad", self.processGamepad)
        self.core.subscribe("/heading", self.processHeading)
        self.core.registerTopic(self.core.getParam(self.PARAM_BEARING_TOPIC), MagnetoMessage)
        self.core.subscribe(self.core.getParam(self.PARAM_BEARING_TOPIC), self.processBearing)
        self.core.subscribe("/nav/onPoint", self.processOnPoint)
        self.core.subscribe("/taskProcessor/complete", self.tasksComplete)
        self.core.subscribe(self.STOP_MESSAGE, self.processStop)
        self.core.subscribe("/obstacleAvoidance/detected", self.obstacleDetected)
        self.core.subscribe("/obstacleAvoidance/passed", self.obstaclePassed)

    def obstacleDetected(self, message: EmptyMessage):
        self.obstacle = True

    def obstaclePassed(self, message: EmptyMessage):
        self.obstacle = False

    def processStop(self, message: ValueMessage):
        self.enable = False

    def processBearing(self, message: MagnetoMessage):
        self.bearing = message.heading
        self.updatePid()

    def processHeading(self, message: MagnetoMessage):
        self.heading = message.heading
        self.updatePid()

    def updatePid(self):
        self.pid.update(AngleCalc.calcDifference(self.heading, self.bearing))
        self.processControl()

    def processOnPoint(self, message: EmptyMessage):
        self.waitingForTasks = True
        self.processControl()

    def tasksComplete(self, message: EmptyMessage):
        self.waitingForTasks = False
        self.processControl()

    def resetPid(self, message):
        self.pid = PID(
            self.core.getParam("/DirectionController/pid/p"),
            self.core.getParam("/DirectionController/pid/i"),
            self.core.getParam("/DirectionController/pid/d")
        )
        self.pid.SetPoint = 0

    def pid(self, message: EmptyMessage):
        msg = KvMessage()
        msg.key = 0
        msg.value = self.core.getParam("/controlEnginesPid/p")
        msg.topic = "/control/engines/pid"
        self.core.publish(msg)
        msg.key = 1
        msg.value = self.core.getParam("/controlEnginesPid/i")
        self.core.publish(msg)
        msg.key = 2
        msg.value = self.core.getParam("/controlEnginesPid/d")
        self.core.publish(msg)

    def processGamepad(self, message: GamepadMessage):
        self.direction = message.rax
        self.speed = message.lay
        if self.softOffBtn.encode(message.b1):
            self.speedGamepadOverride = False
        if self.softOffBtn.encode(message.b2):
            self.speedGamepadOverride = True
        if self.manualOnBtn.encode(message.b3):
            self.manual = True
        if self.manualOffBtn.encode(message.b4):
            self.manual = False
        self.processControl()

    def processControl(self):
        if self.manual:
            self.processManual()
        else:
            self.processAutopilot()

    def processManual(self):
        self.drive(self.speed, self.direction)

    def processAutopilot(self):
        turn = self.pid.output * -1
        speed = 90
        if self.waitingForTasks:
            self.drive()
            return
        if self.speedGamepadOverride:
            speed = self.speed
        else:
            if self.obstacle:
                speed = self.core.getParam(self.OBSTACLE_SPEED)
            else:
                speed = self.core.getParam(self.FOLLOW_SPEED)
        self.drive(speed, 90, turn)


    def drive(self, speed = 90, direction = 90, turn = 0):
        l, r, m = self.atd.transform(speed, nav.calc.Calc.toRange(self.direction + turn, 0, 180))
        self.bridgeMessage.left = l
        self.bridgeMessage.right = r
        self.bridgeMessage.enable = self.enable
        self.bridgeMessage.left = int(self.bridgeMessage.left)
        self.bridgeMessage.right = int(self.bridgeMessage.right)
        self.core.publish(self.bridgeMessage)


class AdvancedEnginesController(EnginesControl):
    BEARING_ANGLE = "/followCards/bearing"
    BEARING_ANGLE1 = "/followCards/angle1"
    BEARING_ANGLE2 = "/followCards/angle2"
    BEARING_OFFSET = "/followCards/offset"
    BEARING_OK_ANGLE = "/followCards/okAngle"
    FOLLOW_SPEED = "/followCards/speed"
    OVERRIDE_SPEED = "/followCards/speedOverride"
    DEFAULT_CONTROL = 90
    def __init__(self, core: Core):
        super().__init__(core)
        self.core = core
        self.bridgeMessage = ImprovedBridgeMessage(mode=1)
        self.speed = self.DEFAULT_CONTROL
        self.direction = self.DEFAULT_CONTROL
        self.soft = False
        self.manual = True
        self.inrange = False
        self.pid = None
        self.resetPid(None)
        self.atd = nav.calc.Calc.AckermannToDifferential()
        self.softOnBtn = Button()
        self.softOffBtn = Button()
        self.manualOnBtn = Button()
        self.manualOffBtn = Button()
        self.core.registerTopic("/controlEnginesPid/updatePid", EmptyMessage)
        self.core.subscribe("/controlEnginesPid/updatePid", self.pid)
        self.core.subscribe("/control/engines/fromArduino", self.arduino)
        self.core.subscribe("/gamepad", self.processGamepad)
        self.core.subscribe("/heading", self.processHeading)
        self.core.subscribe("/strL", self.processRfidL)
        self.core.subscribe("/strC", self.processRfidC)
        self.core.subscribe("/strR", self.processRfidR)
        self.bearingAngleParam = self.BEARING_ANGLE1
        self.setParam(self.BEARING_ANGLE, self.BEARING_ANGLE1, 0)
        self.turnDirection = TurnDirection.NONE
        self.turnComplete = True

        self.core.registerTopic("/bearing", MagnetoMessage)
        self.core.subscribe("/bearing", self.processBearing)

    def processBearing(self, message: MagnetoMessage):
        self.core.setParam(self.BEARING_ANGLE, message.heading)

    def processRfidL(self, message: EmptyMessage):
        if self.turnComplete:
            self.soft = False
            self.switchAngle(self.core.getParam(self.BEARING_OFFSET)*-1)
            self.turnDirection = TurnDirection.RIGHT
            self.turnComplete = False

    def processRfidC(self, message: EmptyMessage):
        if self.turnComplete:
            self.soft = False
            self.turnDirection = TurnDirection.NONE
            self.switchAngle(0)
            self.turnComplete = False

    def processRfidR(self, message: EmptyMessage):
        if self.turnComplete:
            self.soft = False
            self.switchAngle(self.core.getParam(self.BEARING_OFFSET))
            self.turnDirection = TurnDirection.LEFT
            self.turnComplete = False

    def resetPid(self, message):
        self.pid = PID(
            self.core.getParam("/DirectionController/pid/p"),
            self.core.getParam("/DirectionController/pid/i"),
            self.core.getParam("/DirectionController/pid/d")
        )

    def arduino(self, message: ImprovedBridgeMessage):
        pass

    def setParam(self, param, param2, offset):
        self.core.setParam(param, self.core.getParam(param2)+offset)

    def switchAngle(self, offset):
        if not self.manual:
            self.speed = self.DEFAULT_CONTROL
        if self.bearingAngleParam == self.BEARING_ANGLE1:
            self.setParam(self.BEARING_ANGLE, self.BEARING_ANGLE2, offset)
            self.bearingAngleParam = self.BEARING_ANGLE2
        else:
            self.setParam(self.BEARING_ANGLE, self.BEARING_ANGLE1, offset)
            self.bearingAngleParam = self.BEARING_ANGLE1

    def pid(self, message: EmptyMessage):
        msg = KvMessage()
        msg.key = 0
        msg.value = self.core.getParam("/controlEnginesPid/p")
        msg.topic = "/control/engines/pid"
        self.core.publish(msg)
        msg.key = 1
        msg.value = self.core.getParam("/controlEnginesPid/i")
        self.core.publish(msg)
        msg.key = 2
        msg.value = self.core.getParam("/controlEnginesPid/d")
        self.core.publish(msg)


    def processGamepad(self, message: GamepadMessage):
        self.direction = message.rax
        self.speed = message.lay
        if self.softOffBtn.encode(message.b1):
            self.soft = False
        if self.softOffBtn.encode(message.b2):
            self.soft = True
        if self.manualOnBtn.encode(message.b3):
            self.manual = True
        if self.manualOffBtn.encode(message.b4):
            self.manual = False

    def processDirection(self, direction: int, turn: int, turnDirection: TurnDirection) -> int:
        if turnDirection == TurnDirection.NONE:
            return nav.calc.Calc.toRange(self.direction + turn, 0, 180)
        if turnDirection == TurnDirection.RIGHT:
            return nav.calc.Calc.toRange(self.direction + (abs(turn)*-1), 0, 180)
        if turnDirection == TurnDirection.LEFT:
            return nav.calc.Calc.toRange(self.direction + abs(turn), 0, 180)
        return 0

    def processHeading(self, message: MagnetoMessage):
        if nav.calc.Calc.inRange(0,  AngleCalc.calcDifference(message.heading, self.core.getParam(self.BEARING_ANGLE)), self.core.getParam(self.BEARING_OK_ANGLE)):
            self.turnComplete = True
            self.turnDirection = TurnDirection.NONE
            if not self.manual:
                self.soft = False
                self.speed = self.core.getParam(self.FOLLOW_SPEED)
            if not self.inrange:
                self.resetPid(None)
                self.inrange = True
            #self.resetPid(None)
            #turn = nav.calc.Calc.calculateTurn(message.heading, self.core.getParam("/followCards/angle1"), self.core.getParam("/DirectionController/turnCoef"), 90)
            self.pid.SetPoint = 0
            self.pid.update(AngleCalc.calcDifference(message.heading, self.core.getParam(self.BEARING_ANGLE)))
            turn = self.pid.output*-1
        else:
            if self.inrange:
                self.resetPid(None)
                self.inrange = False
            self.pid.SetPoint = 0
            self.pid.update(AngleCalc.calcDifference(message.heading, self.core.getParam(self.BEARING_ANGLE)))
            turn = self.pid.output*-1
        if self.manual:
            turn = 0
        #l, r, m = self.atd.transform(self.speed, self.direction)
        self.turnDirection = TurnDirection.NONE
        # ovsp = self.core.getParam(self.OVERRIDE_SPEED)
        # if ovsp != 0:
        #     l, r, m = self.atd.transform(ovsp, self.processDirection(self.direction, turn, self.turnDirection))
        # else:
        l, r, m = self.atd.transform(self.speed, self.processDirection(self.direction, turn, self.turnDirection))
        self.bridgeMessage.left = l
        self.bridgeMessage.right = r
        #print("{} {} {} {} {}".format(self.bridgeMessage.left, self.bridgeMessage.right, self.speed, self.direction + turn, self.enable))
        self.bridgeMessage.enable = self.enable
        if self.soft:
            if self.bridgeMessage.left < 0:
                self.bridgeMessage.left = 0
            if self.bridgeMessage.right < 0:
                self.bridgeMessage.right = 0
        self.bridgeMessage.left = int(self.bridgeMessage.left)
        self.bridgeMessage.right = int(self.bridgeMessage.right)
        self.core.publish(self.bridgeMessage)


class EnginesController:

    def __init__(self, core: Core, diffCon: DifferentialEnginesControl):
        self.core = core
        self.message = EnginesControlMessage()
        self.directionController = DirectionController(core, self, diffCon)
        self.speedController = SpeedController(core, self)

    def controlSpeed(self, speed):
        self.message.setSpeed(speed)
        self.core.publish(self.message)

    def controlDirection(self, direction):
        self.message.setDirection(direction)
        self.core.publish(self.message)

    def lockAngle(self, bearing):
        nav.calc.Calc.calculateTurn()


class DirectionController:
    TURN_COEF_PARAM = "/engines/turnCoef"
    RESET_PID_TOPIC = "/DirectionController/resetPid"
    DIRECTION_CONTROLLER_TURN_TOPIC = "/DirectionController/turn"
    DIRECTION_CONTROLLER_MODE_TOPIC = "/DirectionController/mode"

    def __init__(self, core: Core, controller: EnginesController, diffCon: DifferentialEnginesControl):
        self.core = core
        self.controller = controller
        self.turnCoef = 1
        self.bearing = 0
        self.mode = DirectionMode.OFF
        self.message = None
        self.speed = 90
        self.pid = None
        self.resetPid(None)
        self.core.registerTopic(self.RESET_PID_TOPIC, EmptyMessage)
        self.core.subscribe(self.RESET_PID_TOPIC, self.resetPid)
        self.core.registerTopic(self.DIRECTION_CONTROLLER_MODE_TOPIC, KeyValueMessage)
        self.core.subscribe(self.DIRECTION_CONTROLLER_MODE_TOPIC, self.modeCallback)
        self.regulator = nav.calc.Calc.TimeRegulator(core, "/DirectionController/TimeRegulator", diffCon)

    def resetPid(self, message):
        self.pid = PID(
            self.core.getParam("/DirectionController/pid/p"),
            self.core.getParam("/DirectionController/pid/i"),
            self.core.getParam("/DirectionController/pid/d")
        )

    def modeCallback(self, message: KeyValueMessage):
        mode = DIRECTION_MODE_DICT.get(message.key)
        if mode is not None:
            self.setMode(mode, message.value)

    def publishHeading(self, message: MagnetoMessage):
        val = self.calcTurn(message.heading, self.bearing)
        self.core.publish(ValueMessage(self.DIRECTION_CONTROLLER_TURN_TOPIC, val))
        val = val + 90
        self.controller.controlDirection(val)

    def calcTurn(self, heading, bearing):
        if self.core.getParam("/DirectionController/turnMode") == 0:
            return nav.calc.Calc.calculateTurn(heading, bearing, self.core.getParam("/DirectionController/turnCoef"), 90)
        if self.core.getParam("/DirectionController/turnMode") == 1:
            #self.pid.setpoint = bearing
            self.pid.SetPoint = bearing
            #return self.pid(heading)
            self.pid.update(heading)
            return self.pid.output
        if self.core.getParam("/DirectionController/turnMode") == 2:
            self.regulator.targetValue = bearing
            coef = self.regulator.update(heading)
            return nav.calc.Calc.calculateTurn(heading, bearing, coef, 90)

        # if self.pid is None:
        #     return nav.calc.Calc.calculateTurn(heading, bearing, self.turnCoef, 90)
        # else:
        #     self.pid.setpoint = bearing
        #     return self.pid(heading)

    def publishGamepad(self, message: GamepadMessage):
        self.controller.controlDirection(message.rax)

    def unsub(self):
        self.core.unsubscribe(self.publishHeading)
        self.core.unsubscribe(self.publishGamepad)

    def setMode(self, mode: DirectionMode, value):
        if mode == DirectionMode.HOLD_ANGLE:
            self.bearing = value
            self.unsub()
            self.core.subscribe(MagnetoMessage().topic, self.publishHeading)
        if mode == DirectionMode.GAMEPAD_INPUT:
            self.unsub()
            self.core.subscribe(GamepadMessage().topic, self.publishGamepad)


class SpeedController:
    SLOW_SPEED_PARAM = "/engines/slowSpeed"
    SPEED_CONTROLLER_MODE_TOPIC = "/SpeedController/mode"
    DEFAULT_SPEED = 90

    def __init__(self, core: Core, controller: EnginesController):
        self.core = core
        self.controller = controller
        self.encoders = self.initEncoders(core)
        self.mode = SpeedMode.OFF
        self.speed = self.DEFAULT_SPEED
        self.ticks = 0
        self.distance = 0
        self.core.registerTopic(self.SPEED_CONTROLLER_MODE_TOPIC, KeyValueMessage)
        self.core.subscribe(self.SPEED_CONTROLLER_MODE_TOPIC, self.modeCallback)

    def initEncoders(self, core):
        return [EncoderInfoNode(core, "/encoder/lp", callback=self.encoderUpdate),
                EncoderInfoNode(core, "/encoder/pp", callback=self.encoderUpdate),
                EncoderInfoNode(core, "/encoder/lz", callback=self.encoderUpdate),
                EncoderInfoNode(core, "/encoder/pz", callback=self.encoderUpdate)]

    def modeCallback(self, message: KeyValueMessage):
        mode = SPEED_MODE_DICT.get(message.key)
        if mode is not None:
            self.setMode(mode, message.value)

    def encoderUpdate(self, encoder: EncoderInfoNode):
        avgTicks = statistics.mean(extract(self.encoders, lambda x: x.lastTicks))
        if self.mode == SpeedMode.REACH_DISTANCE or self.mode == SpeedMode.DISTANCE_SPEED_TICKS:
            self.distanceCheck(avgTicks)
        if self.mode == SpeedMode.HOLD_SPEED_TICKS or self.mode == SpeedMode.DISTANCE_SPEED_TICKS:
            self.speed = self.regulator(self.speed, self.ticks, avgTicks)
            self.controller.controlSpeed(self.speed)

    def distanceCheck(self, avgTicks):
        self.distance = self.distance - avgTicks
        if self.distance <= 0:
            self.stop()

    @staticmethod
    def regulator(speed, requiredTicks, inputTicks, coef):
        diff = abs(requiredTicks - inputTicks)
        diff = diff * coef
        if inputTicks > requiredTicks:
            speed = speed - diff
        if inputTicks < requiredTicks:
            speed = speed + diff
        return speed

    def stop(self):
        self.unsub()
        self.mode = SpeedMode.OFF
        self.controller.controlSpeed(self.DEFAULT_SPEED)

    def publishGamepad(self, message: GamepadMessage):

        self.controller.controlSpeed(message.lay)

    def unsub(self):
        self.core.unsubscribe(self.publishGamepad)

    def setMode(self, mode: SpeedMode, value):
        self.unsub()
        self.mode = mode
        if mode == SpeedMode.GAMEPAD_INPUT:
            self.core.subscribe(GamepadMessage().topic, self.publishGamepad)
        if mode == SpeedMode.OFF:
            self.controller.controlSpeed(self.DEFAULT_SPEED)
        if mode == SpeedMode.HOLD_SPEED:
            self.controller.controlSpeed(value)
        if mode == SpeedMode.SPEED_SLOW:
            self.controller.controlSpeed(110)
            #self.controller.controlDirection(core.getParam(self.SLOW_SPEED_PARAM))
        if mode == SpeedMode.HOLD_SPEED_TICKS:
            self.ticks = value
            self.speed = self.DEFAULT_SPEED
        if mode == SpeedMode.REACH_DISTANCE:
            self.distance = value
        if mode == SpeedMode.DISTANCE_SPEED_TICKS:
            self.ticks = value[0]
            self.distance = value[1]
            self.speed = self.DEFAULT_SPEED




if __name__ == '__main__':
    # import time
    # import random
    # core = Core()
    # core.setDebug(False)
    # ec = EnginesController(core)
    # ec.directionController.setMode(DirectionMode.HOLD_ANGLE, 100)
    # dif = DifferentialEnginesControl(core)
    # bm = TestSubscriber(core, BridgeMessage().topic, True)
    # sub = TestSubscriber(core, EnginesControlMessage().topic, True)
    # sub2 = TestSubscriber(core, MagnetoMessage().topic, False)
    # sub3 = TestSubscriber(core, GamepadMessage().topic, False)
    # while True:
    #     m = MagnetoMessage()
    #     m.heading = random.randint(90, 110)
    #     core.publish(m)
    #     time.sleep(1)
    #     m = GamepadMessage()
    #     m.lax = random.randint(90, 110)
    #     core.publish(m)
    #     time.sleep(1)
    core = Core()
    a = AdvancedEnginesController(core)
    a.speed = 70
    a.direction = 80
    a.soft = True
    a.processHeading(MagnetoMessage())
    print(a.bridgeMessage)
