import math
from core.core import Core
from nav.NavObjects import GeoPoint
from messages.core.Message import ValueMessage
#from engines.DifferentialEnginesControl import DifferentialEnginesControl

EARTH_RADIUS = 6371

def degreesToRadians(degrees):
    return math.pi * degrees / 180.0

def radiansToDegrees(radians):
    return radians * (180.0 / math.pi)

def calculateDistance(start:GeoPoint, end:GeoPoint):
    lat1 = degreesToRadians(start.lat)
    lat2 = degreesToRadians(end.lat)
    deltaLat = degreesToRadians(end.lat - start.lat)
    deltaLon = degreesToRadians(end.lon - start.lon)
    a = math.sin(deltaLat / 2) * math.sin(deltaLat / 2) + math.cos(lat1) * math.cos(lat2) * math.sin(deltaLon / 2) * math.sin(deltaLon / 2)
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    return EARTH_RADIUS * c * 1000

def calculateAngle(start:GeoPoint, end:GeoPoint):
    lat1 = degreesToRadians(start.lat)
    lat2 = degreesToRadians(end.lat)
    deltaLon = degreesToRadians(end.lon - start.lon)
    y = math.sin(deltaLon) * math.cos(lat2)
    x = math.cos(lat1) * math.sin(lat2) - math.sin(lat1) * math.cos(lat2) * math.cos(deltaLon)
    return (radiansToDegrees(math.atan2(y, x)) + 360) % 360

def calculateTurn(heading, bearing, coef, maxOutput):
    temp = calculateDifferenceAngles(heading, bearing)
    temp = temp * coef
    if abs(temp) < maxOutput:
        return temp
    else:
        if temp > 0:
            temp = 360 - abs(temp)
            return temp
        else:
            temp = 360 - abs(temp)
            return temp * -1

def calculateDifferenceAngles(heading, bearing):
    return bearing - heading

def inRange(value, range, offset):
    if range + offset > value and value > range - offset:
        return True
    return False

def addSameSign(value, add):
    if value >= 0:
        value = value + add
        return value
    if value < 0:
        value = value + (add*-1)
        return value
    return 0

def inRangeMinMax(value, min, max):
    if min < value and value < max:
        return True
    return False

def toRange(value, min, max):
    if value < min:
        return min
    if value > max:
        return max
    return value

def toByte(value: int) -> int:
    return toRange(value, 0, 255)

def translate(value, leftMin, leftMax, rightMin, rightMax):
    # Figure out how 'wide' each range is
    leftSpan = leftMax - leftMin
    rightSpan = rightMax - rightMin

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(value - leftMin) / float(leftSpan)

    # Convert the 0-1 range into a value in the right range.
    return rightMin + (valueScaled * rightSpan)


class TimeRegulator:

    def __init__(self, core: Core, paramOffset, diffCon):
        """
        Vyzaduje:
        paramOffset/baseCoef
        paramOffset/coefOffset
        paramOffset/coefOffset
        :param core:
        :param paramOffset: cesta ke konfiguraci regulatoru
        """
        self.core = core
        self.paramOffset = paramOffset
        self._lastInput = 0
        self.targetValue = 0
        self.coef = self.getParam("baseCoef")
        self.diffCon = diffCon

    def getParam(self, param):
        return self.core.getParam("{}/{}".format(self.paramOffset, param))

    def update(self, value):
        coefOffset = self.getParam("coefOffset")
        offset = self.getParam("offset")
        #print("Offset {}".format(offset))
        #if inRange(value, self._lastInput, offset):
        if inRange(value, self.targetValue, self.getParam("softOffset")):
            self.diffCon.soft(True)
            self.coef = self.getParam("softCoef")
        else:
            if inRange(value, self._lastInput, offset):
                self.coef += coefOffset
                self.diffCon.soft(False)
            else:
                dm = self.getParam("decreaseMode")
                if dm == 0:
                    self.coef = self.getParam("baseCoef")
                if dm == 1:
                    self.coef -= coefOffset
            self._lastInput = value
            self.coef = toRange(self.coef, self.getParam("min"), self.getParam("max"))
        self.core.publish(ValueMessage("/timeRegulatorCoef", self.coef))
        return self.coef


class ScaleMapper:

    def __init__(self, inMin=0, inMax=0, outMin=0, outMax=0, reverse=False):
        #self.mapper = interp1d([inMin, inMax], [outMin, outMax])
        self.outMax = outMax
        self.outMin = outMin
        self.inMax = inMax
        self.inMin = inMin
        if reverse:
            self.method = self._mapr
        else:
            self.method = self._mapn

    def _mapr(self, value):
        return self.outMax - (self._mapn(value) - self.outMin)

    def _mapn(self, value):
        return translate(value, self.inMin, self.inMax, self.outMin, self.outMax)
        #return self.mapper(value)

    def map(self, value):
        return float(self.method(value))


class AckermannToDifferential:

    def __init__(self):
        self.mapper = ScaleMapper(0, 180, -255, 255)
        self.soft = False

    def transform(self, speed, direction):
        left = right = self.mapper.map(speed)
        dir = self.mapper.map(direction)
        left += dir
        right -= dir
        left = self.corection(left)
        right = self.corection(right)
        return left, right, self.mode(left, right)

    def corection(self, value):
        if self.soft and value < 0:
            return 0
        return value


    def mode(self, left, right):
        if left >= 0 and right >= 0:
            return 2 # dopredu
        if left <= 0 and right > 0:
            return 1 # prava
        if left > 0 and right <= 0:
            return 0 # leva
        if left <= 0 and right <= 0:
            return 3 # dozadu


class PID:

    def __init__(self, kp, ki, kd):
        self.kp = kp
        self.ki = ki
        self.kd = kd
        self.p_error = 0
        self.i_error = 0
        self.d_error = 0

    def update(self, error):
        self.d_error = error - self. p_error
        self.p_error = error
        self.i_error += error

    def calc(self):
        return -self.kp*self.p_error - self.ki*self.i_error - self.kd*self.d_error


class AngleCalculator:

    def __init__(self):
        self._heading = 0
        self._bearing = 0
        self._course = 0
        self._distance = 0
        self._passDistance = 0
        self._left = 0
        self._right = 0

    def update(self, heading):
        self._heading = heading
        return self.process()

    def updateDistance(self, left, right):
        if left == 0 or right == 0:
            return 0
        self._passDistance = (left+right)/2
        return self.process()

    def process(self):
        errorAngle = self._heading - self._course
        restDistance, angle = self.calcAngle(self._passDistance, self._distance, errorAngle)
        self._distance = restDistance
        if errorAngle > 0:
            self._bearing = abs(angle)*-1
        if errorAngle < 0:
            self._bearing = abs(angle)
        return self._bearing

    def setPoint(self, course, distance):
        self._course = course
        self._distance = distance
        self._passDistance = 0
        self._left = 0
        self._right = 0

    def getBearing(self):
        return self._bearing

    def getRestDistance(self):
        return self._distance

    def calcAngle(self, passDistance, fullDistance, errorAngle):
        if passDistance == 0:
            return fullDistance, 0
        restDistance = math.sqrt(fullDistance*fullDistance+passDistance*passDistance-2*(fullDistance*passDistance)*math.cos(degreesToRadians(errorAngle)))
        print("DEB {} {} {} {}".format(restDistance, passDistance, fullDistance, errorAngle))
        return restDistance, radiansToDegrees(math.acos(((restDistance*restDistance)+(passDistance*passDistance)-fullDistance*fullDistance)/(2*restDistance*passDistance)))


class AngleCalc:

    @staticmethod
    def validAngle(angle):
        return inRangeMinMax(angle, -1, 360)

    @staticmethod
    def calcAngle(angle, add):
        if not (AngleCalc.validAngle(angle) or AngleCalc.validAngle(add)):
            raise Exception("Invalid input angle")
        angle += add
        if angle < 0:
            angle = 359 + angle
        if angle > 359:
            angle = angle - 359
        return angle

    @staticmethod
    def calcDifference(sourceAngle, targetAngle):
        a = targetAngle - sourceAngle
        a = (a + 180) % 360 - 180
        return a




if __name__ == '__main__':
    a1 = 350
    a2 = 350
    ac = AngleCalc()

    print(ac.calcDifference(a1, a2))

    # p1 = GeoPoint(lat=49.1, lon=13.1)
    # p2 = GeoPoint(lat=49.1, lon=13.1)
    #
    # print(calculateDistance(p1, p2))
    # print(calculateAngle(p1,p2))

    # ac = AngleCalc()
    # for x in range(0, 360):
    #     print("{} {}".format(x, ac.calcAngle(x,-180)))

    # atd = AckermannToDifferential()
    # atd.soft = False
    # print(atd.transform(110, 89))
    # print(atd.transform(89, 90))
    # print(atd.transform(110, 91))
    # print(atd.transform(110, 0))
    # print(atd.transform(110, 180))
    # print(atd.transform(80, 89))
    # print(atd.transform(80, 91))
    # print("#########")
    # print(atd.transform(90, 90))
    # print(atd.transform(90, 80))
    # print(atd.transform(90, 110))

    # x = AngleCalculator()
    # print(x.calcAngle(50, 100, 90))
    # x.setPoint(90, 100)
    # print("{} {}".format(x.getBearing(), x.getRestDistance()))
    # x.update(93)
    # x.updateDistance(1.3, 1.4)
    # print("{} {}".format(x.getBearing(), x.getRestDistance()))
    # x.update(86)
    # x.updateDistance(2, 2)
    # print("{} {}".format(x.getBearing(), x.getRestDistance()))
    # x.update(90)
    # x.updateDistance(3, 3.4)
    # print("{} {}".format(x.getBearing(), x.getRestDistance()))


