from core.core import Core, Node
from messages.core.Message import Message, EmptyMessage
from enum import Enum
from messages.arduino.KvMessage import KvMessage, KvIntMessage, KvFloatMessage, MessageModeType
from nav.NavObjects import Path, WayPoint
from serialization.SerializationCoreBridge import SerializationCoreBridge
import time
from utils.GeneralUtils import returnDefaultIfNone
from utils.TimeUtils import getDate, getTime, getCurrentTimeMillis
from utils.IOUtils import deserialize, serialize, getPathBase, createDirectories, getDateTimePath
import picamera
from threading import Thread


class CameraPosition(Message):

    def __init__(self, topic = "/cam/position", pan = 90, titl = 90, *args, **kwargs):
        super().__init__(topic, *args, **kwargs)
        self.pan = pan
        self.tilt = titl

    def __str__(self):
        return "Pan: {}\tTilt: {}".format(self.pan, self.titl)


class CameraTaskMode(Enum):
    IMAGE = 1,
    VIDEO = 2


class CameraTask(Message):

    def __init__(self, topic = "/cam/rec", width=800, height=600, mode: CameraTaskMode = CameraTaskMode.IMAGE, count=1, *args, **kwargs):
        super().__init__(topic, *args, **kwargs)
        self.width = width
        self.height = height
        self.mode = mode
        self.count = count
        
    def __str__(self):
        return "Width: {} Height: {} Mode: {} Count: {}".format(self.width, self.height, self.mode, self.count)

class CameraTaskNote:

    def __init__(self, pathName=None, date=None, time=None, path=None, media=None, **kwargs):
        self.pathName = pathName
        self.date = date
        self.time = time
        self.path = path
        self.media = returnDefaultIfNone(media, {})

    def addMedia(self, waypoint, name):
        wp = None
        if waypoint in self.media:
            wp = self.media[waypoint]
        else:
            wp = []
        wp.append(name)
        self.media[waypoint] = wp


class CameraNode(Node):
    REQUIRED_CAMERA_POSITION = "/cam/position"
    REQUIRED_CAMERA_TASK = "/cam/rec"
    REQUIRED_PATH = "/pathOperation/path"
    REQUIRED_DESTINATION_POINT = "/destinationPoint"
    REQUIRED_NAV_COMPLETE = "/nav/complete"
    PROVIDED_CAMERA_TASK_COMPLETE = "/cam/complete"
    BASE_EXPORT_DIRECTORY_PARAM = "/cam/baseDir"
    ARDUINO_TILT_MESSAGE = KvFloatMessage(preamble=123456789, postamble=987654321, key=8, mode=MessageModeType.SET.value)
    ARDUINO_PAN_MESSAGE = KvFloatMessage(preamble=123456789, postamble=987654321, key=9, mode=MessageModeType.SET.value)

    def __init__(self, core: Core, serialBridge: SerializationCoreBridge):
        super().__init__(core)
        self.serialBridge = serialBridge
        self.core.registerTopic(self.REQUIRED_CAMERA_POSITION, CameraPosition)
        self.core.registerTopic(self.REQUIRED_CAMERA_TASK, CameraTask)
        self.core.registerTopic(self.PROVIDED_CAMERA_TASK_COMPLETE, EmptyMessage)
        self.ctn : CameraTaskNote = None
        self.waypointName = None
        self.core.subscribe(self.REQUIRED_CAMERA_POSITION, self.processPosition)
        self.core.subscribe(self.REQUIRED_CAMERA_TASK, self.processCameraTask)
        self.core.subscribe(self.REQUIRED_PATH, self.startNote)
        self.core.subscribe(self.REQUIRED_DESTINATION_POINT, self.processWaypoint)
        self.core.subscribe(self.REQUIRED_NAV_COMPLETE, self.saveNote)
        self.t = None

    def processPosition(self, message: CameraPosition):
        self.t = Thread(target=self.runPosition, args=(message, ))
        self.t.start()

    def runPosition(self, message: CameraPosition):
        print("CAM_POSITION")
        msg = self.ARDUINO_TILT_MESSAGE
        msg.value = message.tilt
        self.serialBridge.write(msg)
        msg = self.ARDUINO_PAN_MESSAGE
        msg.value = message.pan
        self.serialBridge.write(msg)
        time.sleep(2)
        self.core.publish(EmptyMessage(self.PROVIDED_CAMERA_TASK_COMPLETE))

    def processCameraTask(self, message: CameraTask):
        self.t = Thread(target=self.runCameraTask, args=(message, ))
        self.t.start()

    def runCameraTask(self, message: CameraTask):
        print("CAM_TASK")
        with picamera.PiCamera() as camera:
            camera.hflip = True
            camera.resolution = (message.width, message.height)
            camera.start_preview()
            # Camera warm-up time
            time.sleep(3)
            createDirectories(self.ctn.path)
            fileName = str(getCurrentTimeMillis())+".jpg"
            camera.capture(getPathBase(self.ctn.path, fileName))
            self.ctn.addMedia(self.waypointName, fileName)
            camera.close()
        print("CAM_TASK_END")
        time.sleep(2)
        self.core.publish(EmptyMessage(self.PROVIDED_CAMERA_TASK_COMPLETE))

    def startNote(self, message: Path):
        print("CAM_START_NOTE")
        self.ctn = CameraTaskNote(message.name, str(getDate()), str(getTime()), getPathBase(self.core.getParam(self.BASE_EXPORT_DIRECTORY_PARAM),getDateTimePath()))

    def processWaypoint(self, message: WayPoint):
        print("CAM_WAYPOINT")
        self.waypointName = message.name

    def saveNote(self, message: EmptyMessage):
        print("CAM_SAVE_NOTE")
        createDirectories(self.ctn.path)
        filePath = getPathBase(self.ctn.path, "info.json")
        serialize(filePath, self.ctn)
        self.ctn = None


if __name__ == '__main__':
    base = "C:\\Users\\Wulf\\PycharmProjects\\FlaskSocketIOChat\\app"
    tn = CameraTaskNote("Cesta kolem domu", date=str(getDate()), time=str(getTime()), path="media\\2019\\12\\30\\13-55-40")
    
    tn.addMedia("WP1", "1.jpg")
    tn.addMedia("WP1", "2.jpg")
    tn.addMedia("WP2", "1.jpg")

    createDirectories(getPathBase(base, tn.path))
    filePath = getPathBase(base, tn.path, "info.json")
    serialize(filePath, tn)




