from core.core import Node, Core, TestSubscriber
from messages.MessagesUtils import MessagesEqual
from messages.core.Message import EmptyMessage, ValueMessage
from socketio.socketio import FlaskBridge
from utils.GeneralUtils import returnDefaultIfNone
from utils.TimeUtils import getCurrentTimeMillis, getTimeDifference
import threading
import time


class PersistConfigNode(Node):
    REQUIRED_TOPIC = "/core/persistConfig"
    PUBLISHED_TOPIC = "/core/persistConfigDone"

    def __init__(self, core: Core):
        super().__init__(core)
        core.registerTopic(self.REQUIRED_TOPIC, EmptyMessage)
        core.registerTopic(self.PUBLISHED_TOPIC, EmptyMessage)
        core.subscribe(self.REQUIRED_TOPIC, self.persist)

    def persist(self, message: EmptyMessage):
        self.core.persistConfig()
        self.core.publish(EmptyMessage(topic=self.PUBLISHED_TOPIC))


class SendMessageDelayNode(Node):
    REQUIRED_DEFAULT_MESSAGE = "/sendMessageDelay/activate"
    PROVIDED_DEFAULT_MESSAGE = "/sendMessageDelay/elapsed"

    def __init__(self, core: Core):
        super().__init__(core)
        self.message = None
        self.timeout = None
        self.core.registerTopic(self.REQUIRED_DEFAULT_MESSAGE, ValueMessage)
        self.core.registerTopic(self.PROVIDED_DEFAULT_MESSAGE, EmptyMessage)
        self.core.subscribe(self.REQUIRED_DEFAULT_MESSAGE, self.engageMessage)

    def engageMessage(self, message: ValueMessage):
        self.process(EmptyMessage(topic=self.PROVIDED_DEFAULT_MESSAGE), message.value)

    def process(self, message, timeout):
        t = threading.Thread(target=self.sendMessage, args=(message, timeout))
        t.start()

    def sendMessage(self, message, timeout):
        time.sleep(timeout)
        self.core.publish(message)


class UnderControlNode(Node):
    REQUIRED_TOPIC_PONG = "/control/pong"
    PUBLISHED_TOPIC_ENGINES = "/control/engines/enable"
    PUBLISHED_TOPIC_PING = "/control/ping"
    PUBLISHED_TOPIC_DELAY = "/control/delay"

    def __init__(self, core: Core, fb: FlaskBridge):
        super().__init__(core)
        self.fb = fb
        self.turnOffDelay = returnDefaultIfNone(core.getParam("/control/enginesTurnOffDelay"), 2000)
        self.pingOffset = returnDefaultIfNone(core.getParam("/control/pingOffset"), 1)
        self.pingTimeout = returnDefaultIfNone(core.getParam("/control/pingTimeout"), 2000)
        self.core.registerTopic(self.REQUIRED_TOPIC_PONG, EmptyMessage)
        self.core.registerTopic(self.PUBLISHED_TOPIC_PING, EmptyMessage)
        self.core.registerTopic(self.PUBLISHED_TOPIC_DELAY, ValueMessage)
        self.fb.subscribeTopic(self.PUBLISHED_TOPIC_DELAY)
        self.fb.subscribeTopic(self.PUBLISHED_TOPIC_PING)
        self.core.subscribe(self.REQUIRED_TOPIC_PONG, self.processPong)
        self.alive = True
        self.pingTime = 0
        self.waitingForPong = False
        self.t = threading.Thread(target=self.run)
        self.t.start()

    def processPong(self, message: EmptyMessage):
        self.waitingForPong = False
        delay = getTimeDifference(self.pingTime, getCurrentTimeMillis())
        if delay > self.turnOffDelay:
            self.core.publish(ValueMessage(self.PUBLISHED_TOPIC_ENGINES, False))
        self.core.publish(ValueMessage(self.PUBLISHED_TOPIC_DELAY, delay))

    def run(self):
        while self.alive:
            if not self.waitingForPong:
                self.pingTime = getCurrentTimeMillis()
                self.core.publish(EmptyMessage(topic=self.PUBLISHED_TOPIC_PING))
                self.waitingForPong = True
                time.sleep(self.pingOffset)
            else:
                if getTimeDifference(self.pingTime, getCurrentTimeMillis()) > self.pingTimeout:
                    self.processPong(EmptyMessage())
                time.sleep(self.pingOffset)


class Tasks:

    def __init__(self, topic = "/taskProcessor/tasks", tasks = None):
        self.topic = topic
        self.tasks = tasks


class Task:

    def __init__(self, sendMessage, expectedMessage, eqMessagemethod, **kwargs):
        self.sendMessage = sendMessage
        self.expectedMessage = expectedMessage
        self.eqMessageMethod = eqMessagemethod


class TaskProcessor:
    ALL_COMPLETE_TOPIC = "/taskProcessor/complete"

    def __init__(self, core: Core):
        self.core = core
        self.pointer = 0
        self.tasks = None
        self.core.registerTopic(self.ALL_COMPLETE_TOPIC, EmptyMessage)
        self.core.registerTopic(Tasks().topic, Tasks)
        self.core.subscribe(Tasks().topic, self.process)
        self.currentTask = None
        self.t = None

    def process(self, message: Tasks):
        self.pointer = 0
        self.tasks = message.tasks
        self.t = threading.Thread(target=self.next)
        self.t.start()
        # self.next()

    def next(self):
        print("TASK_PROCESSOR {}/{}".format(self.pointer, len(self.tasks)))
        if self.pointer < len(self.tasks):
            task: Task = self.tasks[self.pointer]
            self.pointer += 1
            self.currentTask = task
            self.core.subscribe(task.expectedMessage.topic, self.processInputMessage)
            self.core.publish(task.sendMessage)
        else:
            self.core.publish(EmptyMessage(self.ALL_COMPLETE_TOPIC))
            print("TASK_PROCESSOR_ALL_TASKS_DONE")

    def processInputMessage(self, message):
        print("TASK_PROCESSOR_TASK_DONE")
        if self.currentTask is not None:
            if self.currentTask.eqMessageMethod(message, self.currentTask.expectedMessage):
                self.core.unsubscribe(self.processInputMessage)
                self.t = threading.Thread(target=self.next)
                self.t.start()
                # self.next()


if __name__ == '__main__':
    core = Core()
    m = TaskProcessor(core)
    sm = SendMessageDelayNode(core)
    st = getCurrentTimeMillis()
    # m()
    # m()

    def logEnd(message):
        print(getTimeDifference(st, getCurrentTimeMillis()))

    core.subscribe(m.ALL_COMPLETE_TOPIC, logEnd)
    ts = TestSubscriber(core, "/x", True)
    ts1 = TestSubscriber(core, m.ALL_COMPLETE_TOPIC, True)
    # sm = SendMessageDelayNode(core)
    # sm.process(ValueMessage(topic="/x"), 5)
    # sm.process(ValueMessage(topic="/x"), 5)
    # sm.process(ValueMessage(topic="/x"), 5)
    # sm.process(ValueMessage(topic="/x"), 5)
    
    ts = [
        Task(ValueMessage(sm.REQUIRED_DEFAULT_MESSAGE, 3), EmptyMessage(sm.PROVIDED_DEFAULT_MESSAGE), MessagesEqual.getTopicMessageEqual),
        Task(ValueMessage(sm.REQUIRED_DEFAULT_MESSAGE, 3), EmptyMessage(sm.PROVIDED_DEFAULT_MESSAGE), MessagesEqual.getTopicMessageEqual),
        Task(ValueMessage(sm.REQUIRED_DEFAULT_MESSAGE, 3), EmptyMessage(sm.PROVIDED_DEFAULT_MESSAGE), MessagesEqual.getTopicMessageEqual),
        Task(ValueMessage(sm.REQUIRED_DEFAULT_MESSAGE, 0), EmptyMessage(sm.PROVIDED_DEFAULT_MESSAGE), MessagesEqual.getTopicMessageEqual)
    ]
    core.publish(Tasks(tasks=ts))
    # print("Waiting ...")
    #
    # print("Complete")
    core.stop()