import time

from core.core import Core, Node
from messages.GpsMessage import GpsMessage
from messages.MagnetoMessage import MagnetoMessage
from messages.core.Message import ValueMessage, EmptyMessage
from nav.NavObjects import Path, WayPoint
from nodes.General import PersistConfigNode
from socketio import socketio
from socketio.socketio import FlaskBridge, FlaskBridgeParamNode
from utils.ConvertUtils import convertToWaypoint
import utils.IOUtils
import utils.JsonUtils
import jsonpickle
import random
from nav.calc.Calc import calculateDistance


class NavPathPreprocessor:

    @staticmethod
    def process(path: Path, minimalDistance) -> Path:
        if len(path.waypoints) > 0:
            output = Path(topic=path.topic)
            print("Preprocessing: {} Input: {} Output: {}".format(minimalDistance, str(len(path.waypoints)), str(len(output.waypoints))))
            output.addWaypoint(path.waypoints[0])
            last = path.waypoints[0]
            for wp in path.waypoints:
                wp: WayPoint
                if calculateDistance(last, wp) > minimalDistance:
                    output.addWaypoint(wp)
                    last = wp
            print("Preprocessing: {} Input: {} Output: {}".format(minimalDistance,str(len(path.waypoints)),str(len(output.waypoints))))
            return output



class PathsWrapper(object):

    def __init__(self, paths=None, **kwargs):
        self.paths = paths


class GpsPathOperationNode(Node):
    NODE_PREFIX = "/pathOperation"
    PATH_JSON = "paths.json"
    REQUIRED_TOPIC_START_RECORD = NODE_PREFIX+"/startRecord"
    REQUIRED_TOPIC_END_RECORD = NODE_PREFIX+"/endRecord"
    REQUIRED_TOPIC_GET_PATH = NODE_PREFIX+"/getPath"
    REQUIRED_TOPIC_GPS = "/gps"

    PUBLISHED_TOPIC_PATH = NODE_PREFIX+"/path"
    PUBLISHED_TOPIC_LAST_OPERATION = NODE_PREFIX+"/lastOperation"

    PARAM_PREPROCESS_MINIMAL_DISTANCES = "/nav/minimalDistances"

    def load(self):
        pw = jsonpickle.decode(utils.IOUtils.readUtf8(self.PATH_JSON))
        self.paths = pw.paths
        # print(self.paths)

    def save(self):
        pw = PathsWrapper()
        pw.paths = self.paths
        utils.IOUtils.writeUtf8IfDifferent(self.PATH_JSON, jsonpickle.encode(pw))

    def sendLastOperation(self, message):
        self.core.publish(ValueMessage(self.PUBLISHED_TOPIC_LAST_OPERATION, message))

    def __init__(self, core: Core, fb: FlaskBridge):
        super().__init__(core)
        self.lastOperation = "Init"
        self.fb = fb
        self.rec : Path = None
        try:
            self.load()
        except:
            self.paths = {}
        self.sendLastOperation("Init")

        self.core.registerTopic(self.PUBLISHED_TOPIC_PATH, Path, json=socketio.JSON_MODE_PICKLE)
        self.core.registerTopic(self.PUBLISHED_TOPIC_LAST_OPERATION, ValueMessage)
        self.core.registerTopic(self.REQUIRED_TOPIC_START_RECORD, ValueMessage)
        self.core.registerTopic(self.REQUIRED_TOPIC_END_RECORD, EmptyMessage)
        self.core.registerTopic(self.REQUIRED_TOPIC_GET_PATH, ValueMessage)

        self.fb.subscribeTopic(self.PUBLISHED_TOPIC_PATH)
        self.fb.subscribeTopic(self.PUBLISHED_TOPIC_LAST_OPERATION)

        self.core.subscribe(self.REQUIRED_TOPIC_START_RECORD, self.startRecord)
        self.core.subscribe(self.REQUIRED_TOPIC_END_RECORD, self.endRecord)
        self.core.subscribe(self.REQUIRED_TOPIC_GET_PATH, self.getPath)
        self.core.subscribe(self.REQUIRED_TOPIC_GPS, self.processGps)

    def startRecord(self, message: ValueMessage):
        self.rec = Path(name=message.value)
        self.sendLastOperation("Start recording")

    def endRecord(self, message: EmptyMessage):
        if self.rec is not None:
            self.paths[self.rec.name] = self.rec
            self.rec = None
            self.save()
            self.sendLastOperation("Stop recording")

    def processGps(self, message: GpsMessage):
        if self.rec is not None:
            self.rec.addWaypoint(convertToWaypoint(message, topic=None, name="".join(("WP ", str(len(self.rec.waypoints))))))

    def getPath(self, message: ValueMessage):
        if message.value in self.paths:
            msg = self.paths[message.value]
            msg.topic = self.PUBLISHED_TOPIC_PATH
            tmp = NavPathPreprocessor.process(msg, self.core.getParam(self.PARAM_PREPROCESS_MINIMAL_DISTANCES))
            self.core.publish(tmp)
            self.sendLastOperation("Dotaz na trasu: {}".format(message.value))
        else:
            self.sendLastOperation("Trasa: {} neexistuje".format(message.value))

    def export(self):
        output = ""
        for pathx in self.paths.values():
            output += "{}\n".format(pathx.name)
            for wp in pathx.waypoints:
                wp: WayPoint
                output += "{},{},{}\n".format(wp.name,wp.lat, wp.lon)
            output += "\n\n"
        return output


if __name__ == '__main__':
    core = Core()
    fb = FlaskBridge(core)
    fb.subscribeTopic("/position/rover")
    fb.subscribeTopic("/position/roverHeading")
    fb.subscribeTopic("/position/roverBearing")
    g = GpsPathOperationNode(core, fb)

    topics = [
        "/nav/okdistance",
        "/nav/minimalDistances",
    ]
    pcn = PersistConfigNode(core)
    fbpn = FlaskBridgeParamNode(core, fb, topics)

    print(g.paths["P2"].waypoints[0])
    latr = ( 49.2, 49.3)
    lonr = (13.6,13.7 )
    angl = (60, 90)
    while True:
        core.publish(MagnetoMessage(topic="/position/roverHeading", heading=random.uniform(angl[0], angl[1])))
        core.publish(MagnetoMessage(topic="/position/roverBearing", heading=random.uniform(angl[0], angl[1])))
        core.publish(GpsMessage(topic="/position/rover", lat=random.uniform(latr[0], latr[1]), lon=random.uniform(lonr[0],lonr[1])))
        core.publish(GpsMessage(topic="/gps", lat=random.uniform(latr[0], latr[1]), lon=random.uniform(lonr[0], lonr[1])))
        time.sleep(1)
